import numpy as np
from transforms3d.affines import compose
from transforms3d.euler import euler2mat
from resize.affine import update_affine


def test_affine():
    original_affine = np.eye(4)
    new_affine = update_affine(original_affine, [1, 1, 4])
    expected_diff = np.zeros((4, 4))
    expected_diff[2, 2] = 3.0
    expected_diff[2, 3] = 1.5
    assert np.allclose(new_affine - original_affine, expected_diff)

    T = np.array([14.0, 3.8, -39.1])
    R = euler2mat(np.pi / 16, -np.pi / 16, np.pi / 16)
    Z = np.array([0.8, 0.8, 1.0])
    S = np.array([0.0, 0.0, 0.0])
    original_affine = compose(T, R, Z, S)
    new_affine = update_affine(original_affine, [1, 1, 3])
    expected_diff = np.zeros((4, 4))
    expected_diff[:, 2] = [-0.29921, -0.45734, 1.92388, 0.0]
    expected_diff[:, 3] = [0.1496, 0.22867, 0.96194, 0.0]
    assert np.allclose(new_affine - original_affine, expected_diff, atol=1e-5)

    print("successful")


if __name__ == "__main__":
    test_affine()



"""Resize with correct sampling step implemented with SciPy.

"""
import numpy as np
from scipy.ndimage import map_coordinates

from .abstract import Resize


def resize(
    image,
    dxyz,
    same_fov=True,
    target_shape=None,
    order=3,
    return_coords=False,
):
    """Wrapper function to resize an image using SciPy.

    See :class:`ResizeScipy` for more details.

    Args:
        return_coords (bool): Return sampling coordinates if ``True``.

    Returns
    -------
    result : numpy.ndarray
        The interpolated image.
    coords (optional) : numpy.ndarray
        The sampling coordinates of this image.

    """
    resizer = ResizeScipy(
        image,
        dxyz,
        same_fov=same_fov,
        target_shape=target_shape,
        order=order,
    )
    resizer.resize()
    if return_coords:
        return resizer.result, resizer.coords
    else:
        return resizer.result


class ResizeScipy(Resize):
    """Resizes the image with sampling steps dx, dy, and dz using SciPy.

    Same FOV mode:

    .. code-block::

        |   x   |   x   |   x   |
        | x | x | x | x | x | x |

    Aligning first point mode:

    .. code-block::

        |   x   |   x   |   x   |
          | x | x | x | x | x |

    Note:
        "Replication" padding is used in the interpolation.

    Args:
        image (numpy.ndarray): The image to resample.
        dxyz (tuple[float]): The sampling steps. Less than 1 for upsampling. For
            example, ``(2.0, 0.8)`` for a 2D image and ``(1.3, 2.1, 0.3)`` for
            a 3D image.
        same_fov (bool): Keep the same FOV as possible when ``True``. Otherwise,
            align the first points along each dimension between the input and
            resulting images.
        target_shape (tuple[int]): The target spatial shape if not ``None``. If
            ``same_fov`` is ``True``, additional sizes are symmetrically padded
            at/cropped from both sides of each spatial dimension. If
            ``same_fov`` is ``False``, additional sizes are padded at/cropped
            from the end of each spatial dimension.
        order (int): B-spline interpolation order.

    """

    def __init__(
        self,
        image,
        dxyz,
        same_fov=True,
        target_shape=None,
        order=3,
    ):
        self.order = order
        super().__init__(image, dxyz, same_fov, target_shape)

    def _check_shape(self):
        super()._check_shape()
        assert len(self.image.shape) == len(self.dxyz)

    @property
    def coords(self):
        """Returns the sampling coordinates.

        The shape of the coordinates array is ``(M, N)``. ``N`` is the number of
        points to sample. It is equal to product of the shape of the resulting
        image :meth:`result`. ``M`` is the dimension of these sampled points
        (e.g., 1, 2, 3). Each column is the coordinates vector of a point to
        sample relative to the input image.

        """
        return self._coords

    def _format_coords(self):
        self._coords = np.meshgrid(*self._coords, indexing="ij")
        self._coords = np.array([c.flatten() for c in self._coords])

    def _resize(self):
        self._result = map_coordinates(
            self.image,
            self.coords,
            mode="nearest",
            order=self.order,
        )
        self._result = self._result.reshape(self._new_shape)



import numpy as np


class Resize:
    """Resizes the image with sampling steps dx, dy, and dz.

    Same FOV mode:

    .. code-block::

        |   x   |   x   |   x   |
        | x | x | x | x | x | x |

    Aligning first point mode:

    .. code-block::

        |   x   |   x   |   x   |
          | x | x | x | x | x |

    """

    def __init__(
        self,
        image,
        dxyz,
        same_fov=True,
        target_shape=None,
    ):
        self.image = image
        self.dxyz = dxyz
        self.same_fov = same_fov
        self.target_shape = target_shape

        self._old_shape = self.image.shape
        self._result = None
        self._coords = None
        self._check_shape()

    def _check_shape(self):
        if self.target_shape is not None:
            assert len(self.target_shape) == len(self.dxyz)

    def resize(self):
        """Call this method to resize the input ``image``."""
        self._calc_sampling_coords()
        self._format_coords()
        self._resize()

    @property
    def result(self):
        """Returns the interpolated result."""
        return self._result

    @property
    def coords(self):
        """Returns the sampling coordinates."""
        return self._coords

    def _calc_sampling_coords(self):
        if self.same_fov:
            self._old_fov = self._calc_old_fov_same_fov()
            self._new_shape = self.target_shape
            if self.target_shape is None:
                self._new_shape = self._calc_new_shape_same_fov()
            self._new_fov = self._calc_new_fov_same_fov()
            self._coords = self._calc_sampling_coords_same_fov()
        else:
            self._old_fov = self._calc_old_fov_align_first()
            self._new_shape = self.target_shape
            if self.target_shape is None:
                self._new_shape = self._calc_new_shape_align_first()
            self._new_fov = self._calc_new_fov_align_first()
            self._coords = self._calc_sampling_coords_align_first()

    def _format_coords(self):
        raise NotImplementedError

    def _resize(self):
        raise NotImplementedError

    def _calc_sampling_coords_same_fov(self):
        coords = list()
        for l, r, d in zip(self._new_fov[0], self._new_fov[1], self.dxyz):
            coords.append(np.arange(l + d / 2, r - d / 4, d))
        return coords

    def _calc_sampling_coords_align_first(self):
        return [np.arange(0, f + d / 4, d) for f, d in zip(self._new_fov, self.dxyz)]

    def _calc_old_fov_same_fov(self):
        """Calculates the FOV of the original image.

        Suppose the left boundaries are at (-0.5, -0.5, -0.5), then the first
        voxel is at (0, 0, 0). Assume the step size is 1.

        """
        step_size = 1
        left = (-0.5,) * len(self._old_shape)
        right = tuple(l + s * step_size for l, s in zip(left, self._old_shape))
        return left, right

    def _calc_old_fov_align_first(self):
        """Assumes the FOV starts from the first and ends at the last point."""
        step_size = 1
        fov = tuple((s - 1) * step_size for s in self._old_shape)
        return fov

    def _calc_new_fov_same_fov(self):
        """Calculates the FOV of the resulting image.

        Assume the old and new FOV have the same center, then the new FOV is
        shifted from the old FOV by half of the size difference.

        """
        old_size = [r - l for l, r in zip(*self._old_fov)]
        new_size = [s * d for s, d in zip(self._new_shape, self.dxyz)]
        size_diff = [(o - n) / 2 for o, n in zip(old_size, new_size)]
        left = tuple(l + sd for l, sd in zip(self._old_fov[0], size_diff))
        right = tuple(r - sd for r, sd in zip(self._old_fov[1], size_diff))
        return left, right

    def _calc_new_fov_align_first(self):
        """The new FOV also starts from the first point."""
        fov = tuple((s - 1) * d for s, d in zip(self._new_shape, self.dxyz))
        return fov

    def _calc_new_shape_same_fov(self):
        return tuple(int(round(s / d)) for s, d in zip(self._old_shape, self.dxyz))

    def _calc_new_shape_align_first(self):
        """The largest number of points to ensure new FOV is within the old."""
        return tuple(
            int(np.floor((s - 1) / d) + 1) for s, d in zip(self._old_shape, self.dxyz)
        )



#!/usr/bin/env python

import numpy as np
from degrade.degrade import *


def test_fwhm_units_to_voxel_space():

    fwhm_space = 3.3
    voxel_space = 1

    assert fwhm_units_to_voxel_space(fwhm_space, voxel_space) == 3.3

    fwhm_space = 2
    voxel_space = 0.5

    assert fwhm_units_to_voxel_space(fwhm_space, voxel_space) == 4

    fwhm_space = 3
    voxel_space = 1.5

    assert fwhm_units_to_voxel_space(fwhm_space, voxel_space) == 2


def test_std_fwhm_conversion():

    for sigma in [0.3, 1, 1.45]:
        assert np.isclose(fwhm_to_std(std_to_fwhm(sigma)), sigma)

    for fwhm in [0.287, 1, 2.485]:
        assert np.isclose(std_to_fwhm(fwhm_to_std(fwhm)), fwhm)


def test_fwhm_needed():

    hr_fwhm = 1
    lr_fwhm = 1
    assert np.isclose(fwhm_needed(hr_fwhm, lr_fwhm), 0)

    hr_fwhm = 1
    lr_fwhm = 4

    assert np.isclose(fwhm_needed(hr_fwhm, lr_fwhm), 3.8729833)

    hr_fwhm = 0.8
    lr_fwhm = 2.5

    assert np.isclose(fwhm_needed(hr_fwhm, lr_fwhm), 2.368543)



import numpy as np
from transforms3d.affines import decompose, compose


def update_affine(affine, scales):
    """Updates affine matrix to take into account new resolution

    Args:
        affine (numpy.ndarray): The affine matrix to update.
        scales (tuple[float] or list[float]): Resolution scales in each direction.
            Less than 1 for upsampling. For example, ``(2.0, 0.8)`` for a 2D image
            and ``(1.3, 2.1, 0.3)`` for a 3D image.

    """
    # Decompose input affine
    tranforms, rotation, zooms, shears = decompose(affine)

    # Adjust zooms
    zooms_new = zooms * np.array(scales)

    # Calculate translation adjustment
    t_val = rotation.dot(zooms_new / 2 * ((1 / np.array(scales)) - 1))

    # Return the new composed affine matrix
    return compose(tranforms - t_val, rotation, zooms_new, shears)



"""
To degrade a signal is to blur and downsample it.

Blurring is convolution with a kernel. Some design parameters
are provided to calculate the kernel with particular FWHMs.

Most support is currently for either a user-provided kernel or
for a Gaussian kernel.
"""
import numpy as np
import sigpy.mri.rf as rf
from scipy import ndimage
from scipy.signal import windows

from ..resize.scipy import resize

try:
    import torch
    from torch.nn import functional
    from ..resize.pytorch import resize as resize_pytorch
except ImportError:
    torch = None
    functional = None
    resize_pytorch = None


def fwhm_units_to_voxel_space(fwhm_space, voxel_space):
    """
    Translate the spatial resolution of the specified FWHM
    into voxel space. This is done by simply taking the ratio.

    For example, say the FWHM is specified to be 2 micrometers,
    but the resolution of each voxel is 0.5 micrometers. Therefore,
    the corresponding FWHM in 2 microns should span 4 voxels (2 / 0.5)

    Args:
        fwhm_space (float): the physical measurement of the FWHM
        voxel_space (float): the physical measurement of the voxel resolution

    Returns:
        (float): The resultant FWHM in number of voxels
    """
    return fwhm_space / voxel_space


def std_to_fwhm(sigma):
    """
    Convert the standard deviation of a Gaussian kernel to
    its corresponding FWHM.

    Args:
        sigma (float): the standard deviation of the Gaussian kernel

    Returns:
        (float): The corresponding FWHM
    """
    return 2 * np.sqrt(2 * np.log(2)) * sigma


def fwhm_to_std(gamma):
    """
    Convert the FWHM of a Gaussian kernel to
    its corresponding standard deviation.

    Args:
        gamma (float): the FWHM of the Gaussian kernel

    Returns:
        (float): The corresponding standard deviation
    """
    return gamma / (2 * np.sqrt(2 * np.log(2)))


def fwhm_needed(fwhm_hr, fwhm_lr):
    """
    We model the resolution of a signal by the FWHM of the PSF
    at acquisition time.

    When simulating the forward process, we want to "land on" the
    specified resolution; this means we wish our result to have
    the specified FWHM.

    If our PSF is Gaussian, we can directly calculate the FWHM of the
    blur kernel needed to arrive at the target FWHM. When we convolve
    two Gaussian kernels, we add their variances. Thus, to find the
    correct blur kernel, we can take a difference of the variances between
    our input FWHM and output FWHM.

    Args:
        fwhm_hr (float): The FWHM of the high resolution signal
        fwhm_lr (float): The FWHM of the low resolution signal

    Returns:
        (float): the FWHM of the blur kernel needed to bring the high resolution
                 signal to the desired low resolution
    """
    # First move specified FWHM to std
    std_hr = fwhm_to_std(fwhm_hr)
    std_lr = fwhm_to_std(fwhm_lr)

    # target std is root diff of variances
    std_target = np.sqrt(std_lr**2 - std_hr**2)

    # Translate back to FWHM space
    return std_to_fwhm(std_target)


def pulse_based_profile(
    window_size,
    slice_thickness,
    tb=16,
    num_samples=128,
    d1=0.01,
    d2=0.01,
    ptype="ex",
    ftype="ls",
):
    """
    Physics-based simulation of slice profile via the SLR algorithm.
    First design an RF pulse, then use ABRM to simulate the magnetization,
    and finally convert to a slice profile for use in digital convolution.

    Args:
        window_size (int): number of taps in the slice profile
        slice_thickness (float): thickness of the desired slice in spation metric units (cm, mm, um, etc.)
        tb (int): time-bandwidth product for the RF pulse
        num_samples (int): number of sampels for the RF pulse
        d1 (float): passband ripple level
        d2 (float): stopband ripple level
        ptype (str): pulse type, defaults to \'ex\' [\'st\' (small-tip excitation), \'ex\' (pi/2 excitation pulse),
            \'se\' (spin-echo pulse), \'inv\' (inversion), or \'sat\' (pi/2 saturation pulse)]
        ftype (str): filter type, defaults to \'ls\' [\'ms\' (sinc), \'pm\' (Parks-McClellan equal-ripple),
            \'min\' (minphase using factored pm), \'max\' (maxphase using factored pm), \'ls\' (least squares)]

    Returns:
        (np.array(dtype=np.float32)): A unit-energy 1D numpy array of the simulated slice profile
    """

    # design pulse
    pulse = rf.slr.dzrf(num_samples, tb, ptype, ftype, d1, d2, False)

    # simulate magnetization
    t = np.linspace(-2 * tb, 2 * tb, num_samples * 4)
    # noinspection PyTypeChecker
    [a, b] = rf.sim.abrm(pulse, t, balanced=True)
    mxy = 2 * np.multiply(np.conj(a), b)

    # convert to unit-energy magnitude for digital convolution
    slice_profile = np.abs(mxy)
    slice_profile /= slice_profile.sum()

    # resample s.t. each sample of the kernel is in physical measurements
    space = t * slice_thickness / tb
    dx = space[1] - space[0]
    slice_profile = resize(
        slice_profile, dxyz=(1 / dx,), order=5, target_shape=(window_size,)
    )

    return slice_profile


WINDOW_OPTIONS = [
    "blackman",
    "hann",
    "hamming",
    "gaussian",
    "cosine",
    "parzen",
    "rect",
    "rf-pulse-slr",
    "boxcar",
]


def select_kernel(window_size, window_choice=None, fwhm=None, sym=True):
    """
    Utility function to select a blur kernel.

    Args:
        window_size (int): the number of taps for the kernel
        window_choice (string): the specific shape of the kernel; one of:
            - \'gaussian\'
            - \'hann\'
            - \'hamming\'
            - \'cosine\'
            - \'parzen\'
            - \'blackman\'
            - \'boxcar\'
            - \'rf-pulse-slr\' (simulate an RF pulse and get corresponding slice profile)
        fwhm (float): The FWHM of the kernel
        sym (bool): Whether the kernel should be symmetric

    Returns:
        (np.array): the parameterized kernel as a numpy array
    """

    if window_choice is None:
        window_choice = np.random.choice(WINDOW_OPTIONS)
    elif window_choice not in WINDOW_OPTIONS:
        raise ValueError("Window choice (%s) is not supported." % window_choice)

    if window_choice == "rf-pulse-slr":
        return pulse_based_profile(window_size=window_size, slice_thickness=fwhm)

    # ===== Special case for rect =====
    if window_choice in ["rect", "boxcar"]:
        w = np.zeros(window_size, dtype=np.float32)
        leng = window_size // 2

        for i in range(window_size):
            # shift val up
            n = np.abs(i - leng)
            if n > fwhm / 2:
                w[i] = 0
            elif n == fwhm / 2:
                w[i] = 0.5
            elif n < fwhm / 2:
                w[i] = 1

        w = w / w.sum()
        return w

    # ===== Otherwise use `windows` module =====
    window = getattr(windows, window_choice)
    if window_choice in ["gaussian"]:
        return window(window_size, fwhm_to_std(fwhm), sym)
    else:
        return window(window_size, sym)


def apply_degrade(
    x, hr_res, lr_res, slice_separation, kernel_type, axis, window_size=None, order=3
):
    """
    Degrade a 3D volume to a desired slice thickness and slice separation
    using the specified kernel along the specified axis.
    """
    fwhm = fwhm_units_to_voxel_space(fwhm_needed(hr_res, lr_res), hr_res)

    if not window_size:
        window_size = int(2 * round(fwhm) + 1)

    scales = [1, 1, 1]
    scales[axis] = slice_separation / hr_res

    kernel = select_kernel(
        window_size=window_size, window_choice=kernel_type, fwhm=fwhm
    )
    # Unit energy
    kernel = kernel / kernel.sum()

    # Blur, then resample
    x_lr = ndimage.convolve1d(x, kernel, mode="nearest", axis=axis)
    x_lr = resize(x_lr, dxyz=scales, order=order)

    return x_lr


def blur(x, blur_fwhm, axis, kernel_type="gaussian", kernel_file=None):
    """
    Blur a signal in 1D by convolution with a blur kernel along a
    specified axis. The signal is edge-padded to keep its original size.

    Args:
        x (np.array: shape (N1, N2, ..., NN) OR torch.Tensor of the same shape): signal to be blurred
        blur_fwhm (float): The FWHM of the blur kernel
        axis (int): the axis along which to blur
        kernel_type (string): The shape of the blur kernel
        kernel_file (string): The filepath to a user-specified kernel
                              as a numpy file; must be `.npy`

    Returns:
        (np.array: shape (N1, N2, ..., NN)): The blurred signal
    """
    if kernel_file is not None:
        kernel = np.load(kernel_file)
    else:
        window_size = int(2 * round(blur_fwhm) + 1)
        kernel = select_kernel(window_size, kernel_type, fwhm=blur_fwhm)
    kernel /= kernel.sum()  # remove gain

    if isinstance(x, np.ndarray):
        blurred = ndimage.convolve1d(x, kernel, mode="nearest", axis=axis)
    elif torch and isinstance(x, torch.Tensor):
        # TODO: implementation in PyTorch at the moment only applies a 1D kernel
        # to a 2D image. This needs to be generalized in the future.
        kernel = kernel.squeeze()[None, None, :, None]
        kernel = torch.tensor(kernel).float().to(x.device)

        # TODO: Since we expect to run the kernel on a 2D image, we expect
        # `x` to be of shape (B, C, H, W) already. In the future this needs to be
        # generalized.
        blurred = functional.conv2d(x, kernel, padding="same")
    else:
        raise TypeError("Input signal should be a NumPy array or PyTorch tensor.")
    return blurred


def alias(img, k, order, axis):
    """
    Introduce aliasing in a signal in 1D by downsampling without applying a
    low-pass filter. This is a phenomena which occurs in all signals when a
    sufficient bandwidth low-pass filter is NOT applied to a signal ahead of time.
    So when we downsample an image, we introduce aliasing in the frequency domain,
    which in turn affects the image domain.

    Args:
        img (np.array: shape (N1, N2, ..., NN)): ND array to be aliased
        k (float): The downsampling factor
        order (int): The order of the B-spline used to downsample. Must
                     be in the set {0, 1, 3, 5}
        axis (int): the axis along which to introduce aliasing

    Returns:
        (np.array: shape (N1, N2, ..., NN)): The signal with aliasing artifacts
    """
    dxyz_down = [1.0 for _ in img.shape]
    dxyz_down[axis] = k

    if torch and isinstance(img, torch.Tensor):
        # TODO: Since we expect to run the kernel on a 2D image, we expect
        # `x` to be of shape (B, C, H, W) already. In the future this needs to be
        # generalized.

        dxyz_down = [1.0 for _ in img.shape[2:]]
        dxyz_down[axis] = k

        return resize_pytorch(img, dxyz=dxyz_down, order=order)

    return resize(img, dxyz=dxyz_down, order=order)



import torch

"""Resize with correct sampling step implemented with PyTorch

"""
import torch.nn.functional as F

from .abstract import Resize


def resize(
    image,
    dxyz,
    same_fov=True,
    target_shape=None,
    order=3,
    return_coords=False,
):
    """Wrapper function to resize an image using PyTorch.

    See :class:`ResizeTorch` for more details.

    Args:
        return_coords (bool): Return sampling coordinates if ``True``.

    Returns
    -------
    result : torch.Tensor
        The interpolated images.
    coords (optional) : torch.Tensor
        The sampling coordinates of these images.

    """
    resizer = ResizeTorch(
        image,
        dxyz,
        same_fov=same_fov,
        target_shape=target_shape,
        order=order,
    )
    resizer.resize()
    if return_coords:
        return resizer.result, resizer.coords
    else:
        return resizer.result


class ResizeTorch(Resize):
    """Resizes the image with sampling steps dx, dy, and dz in PyTorch.

    Same FOV mode:

    .. code-block::

        |   x   |   x   |   x   |
        | x | x | x | x | x | x |

    Aligning first point mode:

    .. code-block::

        |   x   |   x   |   x   |
          | x | x | x | x | x |

    Note:
        * "Replication" padding is used in interpolation.
        * For images with 3D spatial shape (5D input shape), only linear
          (``order = 1``) and nearest (``order = 0``) are supported.
        * For images with 2D spatial shape (4D input shape), cubic
          (``order = 3``) interpolation is also supported.
        * Images with 1D spatial shape is not supported.

    Args:
        image (torch.Tensor): The image to resample with shape ``(B, C,
            *spatial_size)``; ``B`` is the batch size, and ``C`` is the number
            of channels.
        dxyz (tuple[float]): The sampling steps. Less than 1 for upsampling. For
            example, ``(2.0, 0.8)`` for images with 2D spatial shape (4D input
            shape) and ``(1.3, 2.1, 0.3)`` for images with 3D spatial shape (5D
            input shape with batch and channel dimensions).
        same_fov (bool): Keep the same FOV as possible when ``True``. Otherwise,
            align the first points along each dimension between the input and
            resulting images.
        order (int): Interpolation order. 0: nearest; 1: linear; 2: cubic. See
            :func:`torch.nn.functional.grid_sample` for more details of
            the interpolation mode.
        target_shape (tuple[int]): The target spatial shape if not ``None``. If
            ``same_fov`` is ``True``, additional sizes are symmetrically padded
            at/cropped from both sides of each spatial dimension. If
            ``same_fov`` is ``False``, additional sizes are padded at/cropped
            from the end of each spatial dimension.

    """

    def __init__(
        self,
        image,
        dxyz,
        same_fov=True,
        target_shape=None,
        order=3,
    ):
        self.order = order
        self._mode = self._get_mode(self.order)
        super().__init__(image, dxyz, same_fov, target_shape)
        self._old_shape = self.image.shape[2:]

    def _get_mode(self, order):
        if order == 3:
            return "bicubic"
        elif order == 1:
            return "bilinear"
        elif order == 0:
            return "nearest"

    def _check_shape(self):
        super()._check_shape()
        assert len(self.image.shape) == len(self.dxyz) + 2

    @property
    def coords(self):
        """Returns the sampling coordinates.

        The shape of the coordinates array is ``(B, *spatial_size, M)`` where
        ``B`` is the batch size, and ``M`` is 3 for 3D spatial size and 2 for 2D
        spatial size. The ``spatial_size`` is the same with the resuling
        interpolated image. All ``B`` images of this batch share the same
        sampling indices.

        """
        return self._coords

    def _format_coords(self):
        # Map into the coordinates into [-1, 1] as required by F.grid_sample
        self._coords = self._normalize_coords()
        self._coords = torch.meshgrid(*self._coords, indexing="ij")
        self._coords = [c[None, ..., None] for c in self._coords]
        # Reverse the order of  coordinates for F.grid_sample
        self._coords.reverse()
        self._coords = torch.cat(self._coords, -1)
        # The first dim of coords is batch size. PyTorch needs it to be the same
        # with the images to sample.
        repeats = [self.image.shape[0]] + [1] * (self._coords.ndim - 1)
        self._coords = self._coords.repeat(repeats)

    def _normalize_coords(self):
        if self.same_fov:
            fov = [
                stop - start - 1
                for start, stop in zip(self._old_fov[0], self._old_fov[1])
            ]
        else:
            fov = self._old_fov
        return [
            torch.tensor(c / f * 2 - 1).to(self.image)
            for c, f in zip(self._coords, fov)
        ]

    def _resize(self):
        self._result = F.grid_sample(
            self.image,
            self._coords,
            mode=self._mode,
            align_corners=True,
            padding_mode="border",
        )



from pathlib import Path
from setuptools import setup, find_namespace_packages


__package_name__ = "radifox-utils"


def get_version_and_cmdclass(pkg_path):
    """Load version.py module without importing the whole package.

    Template code from miniver
    """
    import os
    from importlib.util import module_from_spec, spec_from_file_location

    spec = spec_from_file_location("version", os.path.join(pkg_path, "_version.py"))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.get_cmdclass(pkg_path)


__version__, cmdclass = get_version_and_cmdclass("radifox/utils")

setup(
    name=__package_name__,
    version=__version__,
    cmdclass=cmdclass,
    author="JH-MIPC",
    author_email="jhmipc@jh.edu",
    url="https://github.com/jh-mipc/radifox-utils",
    description="Medical image utilities provided with RADIFOX",
    long_description=(Path(__file__).parent.resolve() / "README.md").read_text(),
    long_description_content_type="text/markdown",
    license="Apache License, 2.0",
    packages=find_namespace_packages(include=["radifox.*"]),
    entry_points={
        "console_scripts": [
            "degrade-2d=radifox.utils.degrade.degrade_2d:main",
            "degrade-3d=radifox.utils.degrade.degrade_3d:main",
        ]
    },
    install_requires=[
        "nibabel",
        "numpy",
        "scipy",
        "sigpy-lite",
        "transforms3d",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pytest-cov",
        ],
        "pytorch": "torch>=1.10.0",
    },
    python_requires=">=3.7",
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



"""
Degrade an MR volume by cropping in k-space. To mitigate ringing, we
apply a Fermi filter.

Supports both 3D acquisitions and 2D acquisitions.
"""
import argparse
import sys
import time
from contextlib import contextmanager
from pathlib import Path

import nibabel as nib
import numpy as np
from ..resize.affine import update_affine


def apply_fermi(x, target_shape):
    """
    Assume x is in the Fourier domain
    """

    for i, dim in enumerate(x.shape):
        t = np.linspace(-1, 1, dim)
        scale = dim / target_shape[i]
        length = dim
        width = 20.0 / length
        fermi = 1.0 / (1.0 + np.exp((np.abs(t) - 1.0 / scale) / width))

        broadcast_idx = [np.newaxis, np.newaxis]
        broadcast_idx.insert(i, slice(None, None))
        broadcast_idx = tuple(broadcast_idx)
        fermi = fermi[broadcast_idx]

        x *= fermi

    return x


def downsample_k_space(x, target_shape):
    omega = np.fft.fftshift(np.fft.fftn(x))
    old_energy = np.prod(omega.real.shape)
    crops = [(a - b) // 2 for a, b in zip(x.shape, target_shape)]
    crops = [c if c > 0 else 0 for c in crops]
    idx = tuple([slice(c, -c) if c != 0 else slice(None, None) for c in crops])
    omega_crop = omega[idx]

    omega_filt = apply_fermi(omega_crop, target_shape)
    new_energy = np.prod(omega_filt.real.shape)
    omega_filt *= new_energy / old_energy

    y = np.fft.ifftn(np.fft.ifftshift(omega_filt))
    y = np.abs(y)
    return y


@contextmanager
def timer_context(label, verbose=True):
    if verbose:
        print(label)
    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        elapsed_time = end_time - start_time
        if verbose:  # Print elapsed time only if verbose is True
            print(f"\tElapsed time: {elapsed_time:.4f}s")


def process(in_fpath, out_fpath, target_res, verbose=False):
    obj = nib.Nifti1Image.load(in_fpath)
    img = obj.get_fdata(dtype=np.float32)

    zooms = list(obj.header.get_zooms())
    # Update the ratio of the new resolution
    scales = [new / old for new, old in zip(target_res, zooms)]

    # find the ratio of pixels
    downsample_shape = [int(np.round(d / s)) for d, s in zip(img.shape, scales)]

    with timer_context(
        f"=== Cropping and filtering in k-space... ===", verbose=verbose
    ):
        img = downsample_k_space(img, target_shape=downsample_shape)
        new_affine = update_affine(obj.affine, scales)
        nib.Nifti1Image(img, affine=new_affine, header=obj.header).to_filename(
            out_fpath
        )


def main(args=None):
    # ===== Read arguments =====
    parser = argparse.ArgumentParser()
    parser.add_argument("--in-fpath", type=Path, required=True)
    parser.add_argument("--out-fpath", type=Path, required=True)
    parser.add_argument("--target-res", type=str, default=None)
    parser.add_argument("--verbose", action="store_true", default=False)

    parsed_args = parser.parse_args(sys.argv[1:] if args is None else args)

    target_res = [float(r) for r in parsed_args.target_res.split("x")]

    for argname in ["in_fpath", "out_fpath"]:
        setattr(parsed_args, argname, getattr(parsed_args, argname).resolve())

    if not parsed_args.in_fpath.exists():
        raise ValueError("Input filepath must exist.")

    for argname in ["out_fpath"]:
        getattr(parsed_args, argname).parent.mkdir(parents=True, exist_ok=True)

    process(
        parsed_args.in_fpath,
        parsed_args.out_fpath,
        target_res=target_res,
        verbose=parsed_args.verbose,
    )

    if parsed_args.verbose:
        print("Done.")


if __name__ == "__main__":
    main()



"""
Create LR-HR pairs at the specified resolution with the specified slice profile.
"""
import argparse
import sys
import time
from contextlib import contextmanager
from pathlib import Path

import nibabel as nib
import numpy as np

from .degrade import apply_degrade
from ..resize.affine import update_affine


@contextmanager
def timer_context(label, verbose=True):
    if verbose:
        print(label)
    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        elapsed_time = end_time - start_time
        if verbose:  # Print elapsed time only if verbose is True
            print(f"\tElapsed time: {elapsed_time:.4f}s")


def remove_slices(x, n, axis, crop_edge):
    if n == 0:
        return x
    crops = [slice(None, None) for _ in x.shape]
    if crop_edge == "major":
        crops[axis] = slice(None, -n)
    elif crop_edge == "minor":
        crops[axis] = slice(n, None)
    elif crop_edge == "center":
        n1 = int(np.floor(n / 2))
        n2 = int(np.ceil(n / 2))
        crops[axis] = slice(n1, -n2)
    return x[tuple(crops)]


def add_slices(x, n, axis, crop_edge):
    if n == 0:
        return x
    pad_major, pad_minor = 0, 0
    if crop_edge == "major":
        pad_major = n
    elif crop_edge == "minor":
        pad_minor = n
    elif crop_edge == "center":
        pad_minor = int(np.floor(n / 2))
        pad_major = int(np.ceil(n / 2))
    pad_values = [(pad_minor, pad_major) if i == axis else (0, 0) for i in range(3)]
    return np.pad(x, pad_values, mode="reflect")


def is_same_after_scale(a, b):
    return a == int(round(int(round(a / b)) * b))


def find_nearest_same(a, b):
    lower = a
    while not is_same_after_scale(lower, b):
        lower -= 1
    lower = a - lower

    upper = a
    while not is_same_after_scale(upper, b):
        upper += 1
    upper = a - upper

    return lower if abs(lower) <= abs(upper) else upper


def simulate_lr(
    fpath,
    slice_profile,
    slice_thickness,
    slice_separation,
    axis,
    out_lr_fpath,
    out_hr_fpath,
    sizing_edge,
    verbose,
):
    with timer_context(f"=== Loading {fpath}... ===", verbose=verbose):
        # noinspection PyTypeChecker
        obj: nib.Nifti1Image = nib.load(fpath)
        affine = obj.affine
        # noinspection PyTypeChecker
        header: nib.Nifti1Header = obj.header
        x = obj.get_fdata(dtype=np.float32)

        orig_res = round(header.get_zooms()[axis], 3)
        target_res = round(slice_thickness, 3)
        sr_factor = round(
            slice_separation / min([round(i, 3) for i in header.get_zooms()]), 3
        )

    if sizing_edge == "none":
        sizing_op = "pass"
        sizing_str = ""
        n = 0
    else:
        n = find_nearest_same(x.shape[axis], sr_factor)

        sizing_op = "pass" if n == 0 else ("crop" if n > 0 else "pad")
        n = abs(n)
        if sizing_edge == "center":
            sizing_str = f"{int(np.floor(n / 2))} minor and {int(np.ceil(n / 2))} major"
        else:
            sizing_str = f"{n} {sizing_edge}"

    if sizing_op == "crop":
        with timer_context(f"=== Removing {sizing_str} slices... ===", verbose=verbose):
            x_sized = remove_slices(x, n, axis, sizing_edge)
    elif sizing_op == "pad":
        with timer_context(f"=== Adding {sizing_str} slices... ===", verbose=verbose):
            x_sized = add_slices(x, n, axis, sizing_edge)
    else:
        x_sized = x

    with timer_context(f"=== Saving HR image... ===", verbose=verbose):
        nib.Nifti1Image(x_sized, affine=affine, header=header).to_filename(out_hr_fpath)

    with timer_context(
        f"=== Degrading with {slice_profile} to {target_res} || {round(slice_separation - target_res, 3)}... ===",
        verbose=verbose,
    ):
        x_lr = apply_degrade(
            x_sized, orig_res, target_res, slice_separation, slice_profile, axis
        )

    with timer_context(f"=== Saving LR image... ===", verbose=verbose):
        scales = [1, 1, 1]
        scales[axis] = slice_separation / orig_res
        new_affine = update_affine(obj.affine, scales)
        nib.Nifti1Image(x_lr, affine=new_affine, header=header).to_filename(
            out_lr_fpath
        )


def main(args=None):
    # ===== Read arguments =====
    parser = argparse.ArgumentParser()
    parser.add_argument("--in-fpath", type=Path, required=True)
    parser.add_argument("--out-hr-fpath", type=Path, required=True)
    parser.add_argument("--out-lr-fpath", type=Path, required=True)
    parser.add_argument("--axis", type=int, default=2)
    parser.add_argument("--slice-thickness", type=float, required=True)
    parser.add_argument("--slice-separation", type=float, required=True)
    parser.add_argument("--verbose", action="store_true", default=False)
    parser.add_argument(
        "--slice-profile",
        type=str,
        default="rf-pulse-slr",
        choices=["rf-pulse-slr", "gaussian"],
    )
    parser.add_argument(
        "--sizing-edge",
        type=str,
        default="major",
        choices=["major", "minor", "center", "none"],
        help=(
            'Whether to crop/pad the "major" or "minor" indices when creating paired HR-LR data. '
            'Choose "center" to center-crop/pad, biasing towards major if odd. '
            'Choose "none" to skip the cropping step.'
        ),
    )

    parsed_args = parser.parse_args(sys.argv[1:] if args is None else args)

    for argname in ["in_fpath", "out_lr_fpath", "out_hr_fpath"]:
        setattr(parsed_args, argname, getattr(parsed_args, argname).resolve())

    if not parsed_args.in_fpath.exists():
        raise ValueError("Input filepath must exist.")

    for argname in ["out_lr_fpath", "out_hr_fpath"]:
        getattr(parsed_args, argname).parent.mkdir(parents=True, exist_ok=True)

    simulate_lr(
        parsed_args.in_fpath,
        parsed_args.slice_profile,
        parsed_args.slice_thickness,
        parsed_args.slice_separation,
        parsed_args.axis,
        parsed_args.out_lr_fpath,
        parsed_args.out_hr_fpath,
        parsed_args.sizing_edge,
        parsed_args.verbose,
    )
    if parsed_args.verbose:
        print("Done.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# pip install https://github.com/shuohan/resize.git
from resize.scipy import resize
from scipy.ndimage import zoom
from torch.nn.functional import interpolate
import torch
import numpy as np

sampling_step = 0.7
x = np.arange(6).astype(float)
ya = resize(x, (sampling_step,), order=1)
print(ya)  # [0.  0.4 1.1 1.8 2.5 3.2 3.9 4.6 5. ]

yb = zoom(x, 1 / sampling_step, order=1, mode="nearest")
print(yb)  # [0.    0.625 1.25  1.875 2.5   3.125 3.75  4.375 5.   ]

xc = torch.from_numpy(x)[None, None, ...]
yc = interpolate(xc, scale_factor=1 / sampling_step, mode="linear")
print(yc.squeeze().numpy())  # [0.   0.55 1.25 1.95 2.65 3.35 4.05 4.75]
